/* $Id: GeoPoint.h,v 1.13 1997/09/25 20:31:48 remek Exp $ */
/*
 * File:       GeoPoint.h
 * Author:     Pawel Roman (proman@npac.syr.edu)
 * Created:    long time ago
 * Revision:   $Revision: 1.13 $
 * Updated:    $Date: 1997/09/25 20:31:48 $
 * Copyright:  Northeast Parallel Architectures Center
 *             at Syracuse University 1997
 */

#ifndef _GEOPOINT_H
#define _GEOPOINT_H

#include "Comm.h"
#include "GISdb.h"

#include <STL/stlstring.h>
#include <Inventor/SbLinear.h>

class GeoPoint
{
   public:
      GeoPoint( int lon_d, int lon_m, int lon_s,
		int lat_d, int lat_m, int lat_s,
		int alt_met );
      GeoPoint( const SbVec3f& gp_vec, const CoordScale& cs );
      GeoPoint( double lon, double lat, double alt = 0.0 );
      GeoPoint( int lon, int lat, int alt );
      GeoPoint( void );

      operator SbVec3f( void ) const;
      operator string( void ) const;
      GeoPoint operator +( const GeoPoint& gp ) const;
      GeoPoint operator -( const GeoPoint& gp ) const;
      bool operator <( const GeoPoint& gp ) const;
      bool operator ==( const GeoPoint& gp ) const;
      SbVec3f getScaledVec( const CoordScale& cs ) const;
      double distance( const GeoPoint& gp ) const;
      string fullString( void ) const;
      string longString( void ) const;
      float getLon( void ) const;
      float getLat( void ) const;
      const GeoPoint& dump( void )const;

   private:
      //number of seconds
      int lon;
      int lat;
      int alt;
};

// {{{ constructors

inline GeoPoint::
GeoPoint( int lon_d, int lon_m, int lon_s,
	  int lat_d, int lat_m, int lat_s,
	  int alt_met )
{
   ACE_TRACE( "(GeoPoint::GeoPoint(d,m,s))" );
   lon = lon_d * 3600 + lon_m * 60 + lon_s;
   lat = lat_d * 3600 + lat_m * 60 + lat_s;
   alt = alt_met;
   ACE_DEBUG(( LM_DEBUG, "GeoPoint:[%d, %d, %d]\n", lon, lat, alt ));
}

inline GeoPoint::
GeoPoint( int nlon, int nlat, int nalt )
   : lon( nlon ), lat( nlat ), alt( nalt )
{
   ACE_TRACE( "(GeoPoint::GeoPoint(i,i,i))" );
   ACE_DEBUG(( LM_DEBUG, "GeoPoint:[%d, %d, %d]\n", lon, lat, alt ));
}

inline GeoPoint::
GeoPoint( double flon, double flat, double falt )
{
   ACE_TRACE( "(GeoPoint::GeoPoint(d,d,d))" );
   lon = rint( flon * 60. );
   lat = rint( flat * 60. );
   alt = rint( falt );
   ACE_DEBUG(( LM_DEBUG, "GeoPoint:[%d, %d, %d]\n", lon, lat, alt ));
}

inline GeoPoint::
GeoPoint( const SbVec3f& gp_vec, const CoordScale& cs )
{
   ACE_TRACE( "(GeoPoint::GeoPoint(SbVec3f,CoordScale))" );
   float flon, flat, falt;
   gp_vec.getValue( flon, falt, flat );
   lon = rint( flon * 3600. / cs.lon_deg_size );
   lat = rint( flat * 3600. / cs.lat_deg_size );
   alt = rint( falt * 100. / cs.alt_meter_size );
   ACE_DEBUG(( LM_DEBUG, "GeoPoint:[%d, %d, %d]\n", lon, lat, alt ));
}

inline GeoPoint::
GeoPoint( void )
{
   ACE_TRACE( "(GeoPoint::GeoPoint())" );
   lat = lon = alt = 0;
}

// }}}

// {{{ operators

//ugly kludge, different treatment of altitudes in operators + & -
inline GeoPoint GeoPoint::
operator +( const GeoPoint& gp ) const
{
   ACE_TRACE( "(GeoPoint::operator +)" );
   return GeoPoint( lon+gp.lon, lat+gp.lat, alt-gp.alt );
}

inline GeoPoint GeoPoint::
operator -( const GeoPoint& gp ) const
{
   ACE_TRACE( "(GeoPoint::operator -)" );
   return GeoPoint( lon-gp.lon, lat-gp.lat, alt+gp.alt );
}

inline bool GeoPoint::
operator <( const GeoPoint& gp ) const
{
   ACE_TRACE( "(GeoPoint::operator <)" );
   return ( lat < gp.lat )&&( lon < gp.lon );
}

inline bool GeoPoint::
operator ==( const GeoPoint& gp ) const
{
   ACE_TRACE( "(GeoPoint::operator ==)" );
   return ( lat == gp.lat )&&( lon == gp.lon );
}

inline GeoPoint::
operator SbVec3f( void ) const
{
   ACE_TRACE( "(GeoPoint::operator SbVec3f)" );
   return SbVec3f( lon / 3600., lat / 3600., alt );
}
// }}}


inline SbVec3f GeoPoint::
getScaledVec( const CoordScale& cs ) const
{
   ACE_TRACE( "(GeoPoint::getScaledVec)" );
   return SbVec3f( lon / 3600. * cs.lon_deg_size,
		   alt / 100. * cs.alt_meter_size,
		   lat / 3600. * cs.lat_deg_size );
}

inline float GeoPoint::
getLat( void ) const 
{
   return lat/3600.0;
}

inline float GeoPoint::
getLon( void ) const
{
   return lon/3600.0;
}

inline double GeoPoint::
distance( const GeoPoint& gp ) const
{
   ACE_TRACE( "(GeoPoint::distance)" );
   return sqrt( ( lat-gp.lat )*( lat-gp.lat ) +
      ( lon-gp.lon )*( lon-gp.lon ) );
}

#endif
