/* $Id: SceneBuilder.C,v 1.14 1997/09/25 20:31:52 remek Exp $ */
/*
 * File         : SceneBuilder.C
 * Author       : Remek Trzaska
 * Created      : Fri Jul 18 11:46:29 1997
 * Revision     : $Revision: 1.14 $
 * Updated      : $Date: 1997/09/25 20:31:52 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 * Purpose:  Create a scene model consisting of GIS tags
 */

#include "SceneBuilder.h"

static const float DEM_Quality = 0.5; //default dem quality, used in SceneBuilder
static const int TEX_Size = 512; //Preferred texture size, not used yet
static const char SB_DefaultScale[] = "12000 12000 3";
static const char* SB_DefaultRoot;

SceneBuilder::
SceneBuilder(GeoPoint a_NW, GeoPoint SE, int aDegSize, int aDumpScene) :
   NW(a_NW), degSize(aDegSize), dumpScene(aDumpScene)
{
   szLon = (NW.getLon()-SE.getLon())*60.0;
   szLat = (NW.getLat()-SE.getLat())*60.0;

   build();
}

void SceneBuilder::
init()
{
   if( !( SB_DefaultRoot = getenv( "GIS_SB_ROOT" ) ) )
   {
      cerr << "Cannot find GIS_SB_ROOT\n";
      exit( 1 );
   }
}

void SceneBuilder::
build()
{
   scene = new SoSelection;
   TRACE_REF(scene);

   //fog
   if( getenv("GIS_FOG") )
   {
      SoEnvironment* fog = new SoEnvironment;
      fog->fogType.setValue(SoEnvironment::FOG);
      fog->fogVisibility.setValue(6000.0);
      fog->fogColor.setValue(0.3, 0.4, 0.6);
      scene->addChild(fog);
   }
   

   //smoothing
   SoShapeHints* hints = new SoShapeHints;
   hints->creaseAngle.setValue(3.1416);
   scene->addChild(hints);

   //root
   putGISTag(scene, "GIS_Root", SB_DefaultRoot);

   //origin
   putGISTag(scene, "GIS_Origin", NW.fullString().c_str() );

   //bounds
   char buf[512];

   sprintf( buf, "%s %s", string( NW ).c_str(),
	    string( NW-GeoPoint( szLon/60, szLon%60, 0, szLat/60, szLat%60, 0, 0 ) ).c_str() );
   
   putGISTag(scene, "GIS_Bounds", buf);

   //scale
   putGISTag(scene, "GIS_Scale", SB_DefaultScale);
   
   putTextures();
   
   putTerrains();

   if(dumpScene)
   {
      SoWriteAction dumpSceneAction;
      dumpSceneAction.getOutput()->openFile(dumpScenePath);
      dumpSceneAction.getOutput()->setBinary(FALSE);
      dumpSceneAction.apply(scene);
      dumpSceneAction.getOutput()->closeFile();
   }
}

void SceneBuilder::
putGISTag(SoGroup* parent, const char* tag, const char* tagValue)
{
   SoInfo* info = new SoInfo;
   info->setName(tag);
   info->string.setValue(tagValue);
   parent->addChild(info);
};

   
void SceneBuilder::
putTextures()
{
   putGISTag(scene, "GIS_Texture", "MAP,Map (scale 1:250 000)");
   putGISTag(scene, "GIS_Texture", "SAT,Satellite Image");
};

void SceneBuilder::
putTerrains()
{
   int countLon = szLon/degSize;
   int countLat = szLat/degSize;
   for(int y = 0; y < countLat; y++)
      for(int x = 0; x < countLon; x++)
      {
	 int cornerLon = NW.getLon()*60-x*degSize;
	 int cornerLat = NW.getLat()*60-y*degSize;
	 char buf[512];
	 //            lon      lat     deg ql texsz
	 sprintf(buf, "%d:%d:%d,%d:%d:%d %d %f %d",
		 cornerLon/60, cornerLon%60, 0,
		 cornerLat/60, cornerLat%60, 0,
		 degSize,
		 DEM_Quality,
		 TEX_Size);
	 putGISTag(scene, "GIS_Terrain", buf);
      }

};

