/* $Id: SoChunk.C,v 1.16 1998/01/05 19:03:31 remek Exp $ */
/*
 * File         : SoChunk.C
 * Author       : Remek Trzaska
 * Created      : Thu Jul 10 12:16:52 1997
 * Revision     : $Revision: 1.16 $
 * Updated      : $Date: 1998/01/05 19:03:31 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "SoChunk.h"
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoFaceSet.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoMaterial.h>

//length factor for terrain extension
static const int lengthFactor = 1.0;

//  constructor

SoChunk::
SoChunk( const GeoPoint& gp, int dSize, float qual, int tSize, float yS,
	 unsigned char aBoundsTagParsed,
	 float amaxWest, float amaxNorth, float amaxEast, float amaxSouth)
   : origin( gp ), demSize( dSize ), quality( qual ), texSize( tSize ),
     yScale( yS), boundsTagParsed( aBoundsTagParsed ), globalWest (amaxWest),
     globalNorth (amaxNorth), globalEast (amaxEast), globalSouth (amaxSouth)
{
   ACE_TRACE( "(SoChunk::SoChunk)" );
   SoGroup* node;
   TRACE_NEW_2( node, SoGroup );
   addChild( node );  // slot for SoTranslation
   addChild( node );  // slot for SoTextureCoordinatePlane
   
   SoSeparator* mainChunk = new SoSeparator;
   addChild( mainChunk );

   mainChunk->addChild( node );  // slot for SoURLTexture2
   mainChunk->addChild( node );  // slot for SoTerrain
   dem = NULL;
   onBoundary = 0;

   if(boundsTagParsed)
   {
      if(origin.getLat() == globalNorth)
	 onBoundary |= NORTH;
      if(origin.getLon() == globalWest)
	 onBoundary |= WEST;
      if(origin.getLat() - demSize/60.0 == globalSouth)
	 onBoundary |= SOUTH;
      if(origin.getLon() - demSize/60.0 == globalEast)
	 onBoundary |= EAST;
   }
   

   if(onBoundary)
   {
      TRACE_NEW_2(extensionGroup, SoGroup);
      addChild( extensionGroup );
      SoTextureCoordinatePlane* texturePlane = new SoTextureCoordinatePlane;
      texturePlane->directionS.setValue(0, 0, 1/(6*512.0));
      texturePlane->directionT.setValue(1/(6*512.0), 0, 0);
      extensionGroup->addChild(texturePlane);
   }
}

// }}}

//  destructor

SoChunk::
~SoChunk( void )
{
   ACE_TRACE( "(SoChunk::~SoChunk)" );
   ACE_DEBUG(( LM_DEBUG, "this=0x%x, ref=%d\n", this, getRefCount() ));
}

// }}}

void SoChunk::
setTranslation( const SbVec3f& trans )
{
   ACE_TRACE( "(SoChunk::setTranslation)" );
   SoTranslation* tr_node;
   TRACE_NEW_2( tr_node, SoTranslation );
   if( ! tr_node )
   {
      ACE_ERROR(( LM_ERROR, "[!](SoChunk::setTranslation) Not enough memory\n" ));
      return;
   }
   ACE_DEBUG(( LM_DEBUG, "trans: [%.2f, %.2f, %.2f]\n", trans[0], trans[1], trans[2] ));
   tr_node->translation = trans;
   replaceChild( 0, tr_node );
}

//  DEMs handling

void SoChunk::
setDEM( const string& loc , int real)
{
   ACE_TRACE( "(SoChunk::setDEM)" );
#ifdef RO_DEMO
   if(!real)
   {
      SoSeparator* fake = new SoSeparator;

      SoCoordinate3* coords = new SoCoordinate3;
      coords->point.set1Value(0, 0, 0, 0);
      coords->point.set1Value(1, 6000, 0, 0);
      coords->point.set1Value(2, 6000, 0, 6000);
      coords->point.set1Value(3, 0, 0, 6000);
      fake->addChild(coords);

      SoIndexedFaceSet* faces = new SoIndexedFaceSet;
      faces->coordIndex.set1Value(0, 0);
      faces->coordIndex.set1Value(1, 2);
      faces->coordIndex.set1Value(2, 1);
      faces->coordIndex.set1Value(3, -1);
      faces->coordIndex.set1Value(4, 0);
      faces->coordIndex.set1Value(5, 3);
      faces->coordIndex.set1Value(6, 2);
      faces->coordIndex.set1Value(7, -1);
      fake->addChild(faces);
      ((SoSeparator*)getChild(2))->replaceChild( 1, fake );
   }
   else
   {
#endif //RO_DEMO
      char tempBuf[10];
      sprintf(tempBuf, "%d/", demSize);
      string fullURL = loc + tempBuf + origin + ".dem";
      TRACE_NEW_2( dem, SoTerrain( ( char* )fullURL.c_str(), quality,
				   yScale, demSize ) );
      if( dem )
      {
	 dem->build();
	 maxAlt = dem->maxAlt;
	 minAlt = dem->minAlt;
      
	 ((SoSeparator*)getChild(2))->replaceChild( 1, dem );
      
	 if(boundsTagParsed && onBoundary)
	    extendDEM();

	 TRACE_UNREF( dem );
      }
#ifdef RO_DEMO
   }
#endif
}

void SoChunk::
extendDEM()
{
   float demdim = dem->dim-1.5;
   
//  sides

//  north

   if(onBoundary & NORTH)
   {
      SoSeparator* extension = new SoSeparator;
      SoCoordinate3* coords = new SoCoordinate3;
      // 1   0
      // 23..n

      //0
      coords->point.set1Value(0, SbVec3f(demdim*sampleDistance,
					 dem->getElev(0, 0)/yScale * 0.95,
					 -lengthFactor*demdim*sampleDistance));
      //1
      coords->point.set1Value(1, SbVec3f(0,
					 dem->getElev(0, dem->dim-1)/yScale * 0.95,
					 -lengthFactor*demdim*sampleDistance));
      //2
      coords->point.set1Value(2, SbVec3f(0,
					 dem->getElev(0, dem->dim-1)/yScale * 0.95,
					 0));
	 
      //loop
      int pointIndex = 3;

      for(int i=dem->dim-10; i>0; i-=10)
      {
	 float xCoord = (dem->dim-1.5)/(-dem->dim + 1.0)*(i-dem->dim + 1.0)
	    * sampleDistance;
	 float elevation = dem->getElev(0, i)/yScale*0.95;
	 float zCoord = -1.0;
	 coords->point.set1Value(pointIndex++,
				 SbVec3f(xCoord, elevation, zCoord));
      };
      
      //n
      coords->point.set1Value(pointIndex,
			      SbVec3f(demdim*sampleDistance,
				      dem->getElev(0, 0)/yScale * 0.95,
				      0));

      extension->addChild(coords);
      SoIndexedFaceSet* faceSet = new SoIndexedFaceSet;

      int vertexIndex = 0;
      
      for(i=2; i<(pointIndex-1)/2; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 1);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }

      faceSet->coordIndex.set1Value(vertexIndex++, 0);
      faceSet->coordIndex.set1Value(vertexIndex++, 1);
      faceSet->coordIndex.set1Value(vertexIndex++, i);
      faceSet->coordIndex.set1Value(vertexIndex++, -1);

      for(; i<pointIndex; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 0);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }
      
      extension->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[NORTH] = extensionGroup->findChild(group);
      
      extensionGroup->addChild(extension);

   }

// }}}

//  east

   if(onBoundary & EAST)
   {
      SoSeparator* extension = new SoSeparator;
      SoCoordinate3* coords = new SoCoordinate3;
      // 2   1
      // 3
      // .
      // n   0

      //0
      coords->point.set1Value(0, SbVec3f((1+lengthFactor)*demdim*sampleDistance,
					 dem->getElev(dem->dim-1, 0)/yScale * 0.95,
					 demdim*sampleDistance));
      //1
      coords->point.set1Value(1, SbVec3f((1+lengthFactor)*demdim*sampleDistance,
					 dem->getElev(0, 0)/yScale * 0.95,
					 0));
      //2
      coords->point.set1Value(2, SbVec3f(demdim*sampleDistance,
					 dem->getElev(0, 0)/yScale * 0.95,
					 0));
	 
      //loop
      int pointIndex = 3;

      for(int i=10; i<dem->dim-1; i+=10)
      {
	 float xCoord = demdim*sampleDistance - 1.0;
	 float elevation = dem->getElev(i, 0)/yScale*0.95;
	 float zCoord = demdim/(dem->dim-1.0)*i*sampleDistance;
	 coords->point.set1Value(pointIndex++,
				 SbVec3f(xCoord, elevation, zCoord));
      };
      //n
      coords->point.set1Value(pointIndex, SbVec3f(demdim*sampleDistance,
					 dem->getElev(dem->dim-1, 0)/yScale * 0.95,
					 demdim*sampleDistance));

      extension->addChild(coords);

      SoIndexedFaceSet* faceSet = new SoIndexedFaceSet;

      int vertexIndex = 0;
      
      for(i=2; i<(pointIndex-1)/2; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 1);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }

      faceSet->coordIndex.set1Value(vertexIndex++, 0);
      faceSet->coordIndex.set1Value(vertexIndex++, 1);
      faceSet->coordIndex.set1Value(vertexIndex++, i);
      faceSet->coordIndex.set1Value(vertexIndex++, -1);

      for(; i<pointIndex; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 0);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }
      
      extension->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[EAST] = extensionGroup->findChild(group);

      extensionGroup->addChild(extension);
   }

// }}}

//  south

   if(onBoundary & SOUTH)
   {
      SoSeparator* extension = new SoSeparator;
      SoCoordinate3* coords = new SoCoordinate3;
      // n..32
      // 0   1

      //0
      coords->point.set1Value(0, SbVec3f(0,
					 dem->getElev(dem->dim-1, dem->dim-1)/yScale * 0.95,
					 (1+lengthFactor)*demdim*sampleDistance));
      //1
      coords->point.set1Value(1, SbVec3f(demdim*sampleDistance,
					 dem->getElev(dem->dim-1, 0)/yScale * 0.95,
					 (1+lengthFactor)*demdim*sampleDistance));
      //2
      coords->point.set1Value(2, SbVec3f(demdim*sampleDistance,
					 dem->getElev(dem->dim-1, 0)/yScale * 0.95,
					 demdim*sampleDistance));
	 
      //loop
      int pointIndex = 3;

      for(int i=10; i<dem->dim-1; i+=10)
      {
	 float xCoord = (demdim-(dem->dim-1.5)/(dem->dim-1.0)*i)
	    * sampleDistance;
	 float elevation = dem->getElev(dem->dim-1, i)/yScale*0.95;
	 float zCoord = demdim*sampleDistance - 1;
	 coords->point.set1Value(pointIndex++,
				 SbVec3f(xCoord, elevation, zCoord));
      };
      //n
      coords->point.set1Value(pointIndex,
			      SbVec3f(0,
				  dem->getElev(dem->dim-1, dem->dim-1)/yScale * 0.95,
				  demdim*sampleDistance));

      extension->addChild(coords);

      SoIndexedFaceSet* faceSet = new SoIndexedFaceSet;
      int vertexIndex = 0;
      
      for(i=2; i<(pointIndex-1)/2; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 1);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }

      faceSet->coordIndex.set1Value(vertexIndex++, 0);
      faceSet->coordIndex.set1Value(vertexIndex++, 1);
      faceSet->coordIndex.set1Value(vertexIndex++, i);
      faceSet->coordIndex.set1Value(vertexIndex++, -1);

      for(; i<pointIndex; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 0);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }

      extension->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[SOUTH] = extensionGroup->findChild(group);

      extensionGroup->addChild(extension);

   }

// }}}

//  west

   if(onBoundary & WEST)
   {
      SoSeparator* extension = new SoSeparator;
      SoCoordinate3* coords = new SoCoordinate3;
      // 0   n
      //     .
      //     3
      // 1   2

      //0
      coords->point.set1Value(0, SbVec3f(-lengthFactor*demdim*sampleDistance,
					 dem->getElev(0, dem->dim-1)/yScale * 0.95,
					 0));
      //1
      coords->point.set1Value(1, SbVec3f(-lengthFactor*demdim*sampleDistance,
					 dem->getElev(dem->dim-1, dem->dim-1)/yScale * 0.95,
					 demdim*sampleDistance));
      //2
      coords->point.set1Value(2, SbVec3f(0,
					 dem->getElev(dem->dim-1, dem->dim-1)/yScale * 0.95,
					 demdim*sampleDistance));
	 
      //loop
      int pointIndex = 3;

      for(int i=dem->dim-10; i>0; i-=10)
      {
	 float xCoord = 1.0;
	 float elevation = dem->getElev(i, dem->dim-1)/yScale*0.95;
	 float zCoord = demdim/(dem->dim-1.0)*i*sampleDistance;
	 coords->point.set1Value(pointIndex++,
				 SbVec3f(xCoord, elevation, zCoord));
      };
      //n
      coords->point.set1Value(pointIndex, SbVec3f(0,
					 dem->getElev(0, dem->dim-1)/yScale * 0.95,
					 0));

      extension->addChild(coords);
      SoIndexedFaceSet* faceSet = new SoIndexedFaceSet;

      int vertexIndex = 0;
      
      for(i=2; i<(pointIndex-1)/2; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 1);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }

      faceSet->coordIndex.set1Value(vertexIndex++, 0);
      faceSet->coordIndex.set1Value(vertexIndex++, 1);
      faceSet->coordIndex.set1Value(vertexIndex++, i);
      faceSet->coordIndex.set1Value(vertexIndex++, -1);

      for(; i<pointIndex; i++)
      {
	 faceSet->coordIndex.set1Value(vertexIndex++, 0);
	 faceSet->coordIndex.set1Value(vertexIndex++, i);
	 faceSet->coordIndex.set1Value(vertexIndex++, i+1);
	 faceSet->coordIndex.set1Value(vertexIndex++, -1);
      }
      
      extension->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[WEST] = extensionGroup->findChild(group);

      extensionGroup->addChild(extension);
   }

// }}}

// }}}
   
//  corners

   //for each corner if both boundaries
   if((onBoundary & (WEST | NORTH)) == (WEST | NORTH))
   {
      //nw
      // 2  1
      // 3  0
      ELEV value = dem->getElev(0, dem->dim-1) * 0.95;
      SoSeparator* dumbRectangle = new SoSeparator;
      SoCoordinate3* coords = new SoCoordinate3;
      coords->point.set1Value(0, SbVec3f(0, value/yScale, 0));
      coords->point.set1Value(1, SbVec3f(0,
					 value/yScale,
					 -sampleDistance*demdim));
      coords->point.set1Value(2, SbVec3f(-sampleDistance*demdim,
					 value/yScale,
					 -sampleDistance*demdim));
      coords->point.set1Value(3, SbVec3f(-sampleDistance*demdim,
					 value/yScale,
					 0));
      dumbRectangle->addChild(coords);
      SoFaceSet* faceSet = new SoFaceSet;
      faceSet->numVertices = 4;
      dumbRectangle->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[WEST|NORTH] = extensionGroup->findChild(group);

      extensionGroup->addChild(dumbRectangle);
   }

      
   if((onBoundary & (EAST | NORTH)) == (EAST | NORTH))
   {
      //ne
      //3  2
      //0  1
      ELEV value = dem->getElev(0, 0) * 0.95;
      SoSeparator* dumbRectangle = new SoSeparator;
      SoCoordinate3* coords = new SoCoordinate3;
      coords->point.set1Value(0, SbVec3f(2*sampleDistance*demdim,
					 value/yScale,
					 0));
      coords->point.set1Value(1, SbVec3f(2*sampleDistance*demdim,
					 value/yScale,
					 -sampleDistance*demdim));
      coords->point.set1Value(2, SbVec3f(sampleDistance*demdim,
					 value/yScale,
					 -sampleDistance*demdim));
      coords->point.set1Value(3, SbVec3f(sampleDistance*demdim,
					 value/yScale,
					 0));
      dumbRectangle->addChild(coords);
      SoFaceSet* faceSet = new SoFaceSet;
      faceSet->numVertices = 4;
      dumbRectangle->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[EAST|NORTH] = extensionGroup->findChild(group);

      extensionGroup->addChild(dumbRectangle);
   }

   if((onBoundary & (EAST | SOUTH)) == (EAST | SOUTH))
   {
      //se
      ELEV value = dem->getElev(dem->dim-1, 0) * 0.95;
      SoSeparator* dumbRectangle = new SoSeparator;

      SoCoordinate3* coords = new SoCoordinate3;
      coords->point.set1Value(0, SbVec3f(2*sampleDistance*demdim,
					 value/yScale,
					 sampleDistance*demdim));
      coords->point.set1Value(1, SbVec3f(sampleDistance*demdim,
					 value/yScale,
					 sampleDistance*demdim));
      coords->point.set1Value(2, SbVec3f(sampleDistance*demdim,
					 value/yScale,
					 2*sampleDistance*demdim));
      coords->point.set1Value(3, SbVec3f(2*sampleDistance*demdim,
					 value/yScale,
					 2*sampleDistance*demdim));
      dumbRectangle->addChild(coords);
      SoFaceSet* faceSet = new SoFaceSet;
      faceSet->numVertices = 4;
      dumbRectangle->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[EAST|SOUTH] = extensionGroup->findChild(group);

      extensionGroup->addChild(dumbRectangle);
   }


   if((onBoundary & (WEST | SOUTH)) == (WEST | SOUTH))
   {
      //sw
      ELEV value = dem->getElev(dem->dim-1, dem->dim-1)*0.95;
      
      SoSeparator* dumbRectangle = new SoSeparator;

      SoCoordinate3* coords = new SoCoordinate3;
      coords->point.set1Value(0, SbVec3f(0,
					 value/yScale,
					 sampleDistance*demdim));
      coords->point.set1Value(1, SbVec3f(-sampleDistance*demdim,
					 value/yScale,
					 sampleDistance*demdim));
      coords->point.set1Value(2, SbVec3f(-sampleDistance*demdim,
					 value/yScale,
					 2*sampleDistance*demdim));
      coords->point.set1Value(3, SbVec3f(0,
					 value/yScale,
					 2*sampleDistance*demdim));
      dumbRectangle->addChild(coords);
      SoFaceSet* faceSet = new SoFaceSet;
      faceSet->numVertices = 4;
      dumbRectangle->addChild(faceSet);

      SoGroup* group = new SoGroup;
      extensionGroup->addChild(group);  //slot for textures;
      textureSlotsIndices[WEST|SOUTH] = extensionGroup->findChild(group);

      extensionGroup->addChild(dumbRectangle);
   }

// }}}

}

// }}}

//  textures handling

void SoChunk::
extendTex(const string& loc)
{
   if(onBoundary & NORTH)
      addSmallTexture(loc,
		      origin + GeoPoint(0, 0, 0, 0, 15, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[NORTH]));
   
   if(onBoundary & EAST)
      addSmallTexture(loc,
		      origin + GeoPoint(0, -15, 0, 0, 0, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[EAST]));
   
   if(onBoundary & SOUTH)
      addSmallTexture(loc,
		      origin + GeoPoint(0, 0, 0, 0, -15, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[SOUTH]));
   
   if(onBoundary & WEST)
      addSmallTexture(loc,
		      origin + GeoPoint(0, 15, 0, 0, 0, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[WEST]));
      
   //for each corner if both boundaries
   if((onBoundary & (WEST | NORTH)) == (WEST | NORTH))
      addSmallTexture(loc,
		      origin + GeoPoint(0, 15, 0, 0, 15, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[WEST|NORTH]));
   if((onBoundary & (EAST | NORTH)) == (EAST | NORTH))
      addSmallTexture(loc,
		      origin + GeoPoint(0, -15, 0, 0, 15, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[EAST|NORTH]));
   if((onBoundary & (EAST | SOUTH)) == (EAST | SOUTH))
      addSmallTexture(loc,
		      origin + GeoPoint(0, -15, 0, 0, -15, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[EAST|SOUTH]));
   if((onBoundary & (WEST | SOUTH)) == (WEST | SOUTH))
      addSmallTexture(loc,
		      origin + GeoPoint(0, 15, 0, 0, -15, 0, 0),
		      extensionGroup->getChild(textureSlotsIndices[WEST|SOUTH]));
   
};

void SoChunk::
addSmallTexture(const string& loc, GeoPoint geoCoord, SoNode* where,
		int rotate)
{
   SoURLTexture2* tex_node;
   char tempBuf[10];
   sprintf(tempBuf, "%d/", demSize);
   
   string fullURL = loc + tempBuf + "thumbs/" + geoCoord + ".ppm";
   TRACE_NEW_2( tex_node, SoURLTexture2( ( char* )fullURL.c_str(), extensionGroup, where, rotate ) );
   if( ! tex_node )
   {
      ACE_ERROR(( LM_ERROR, "[!](SoChunk::switchTexTo) Not enough memory\n" ));
      return;
   }
};


void SoChunk::
switchTexTo( const string& loc )
{
   ACE_TRACE( "(SoChunk::switchTexTo)" );
   SoURLTexture2* tex_node;
   char tempBuf[10];
   sprintf(tempBuf, "%d/", demSize);

   string fullURL = loc + tempBuf + origin + ".ppm";
   TRACE_NEW_2( tex_node,
		SoURLTexture2( ( char* )fullURL.c_str(),
			       (SoSeparator*)getChild(2),
			       ((SoSeparator*)getChild(2))->getChild(0) ) );
   if( ! tex_node )
   {
      ACE_ERROR(( LM_ERROR, "[!](SoChunk::switchTexTo) Not enough memory\n" ));
      return;
   }

   if(boundsTagParsed && onBoundary)
      extendTex(loc);
}

// }}}
