/* $Id: SoMovingObj.h,v 1.11 1997/09/25 20:31:55 remek Exp $ */
/*
 * File         : SoMovingObj.h
 * Author       : Pawel Roman (proman@npac.syr.edu)
 * Created      : Mon Aug 25 11:57:12 1997
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 1997/09/25 20:31:55 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#ifndef _SOMOVINGOBJ_H
#define _SOMOVINGOBJ_H

#include <stlstring.h>
#include <Inventor/nodes/SoShape.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoFont.h>
#include <Inventor/nodes/SoText2.h>

#include "SoActiveTranslation.h"
#include "SoXtWebViewer.h"

class SoMovingObj : public SoSeparator
{
   public:
      enum Class { FRIEND, FOE, UFO };
      enum Type { AIRPLANE, CHOPPER };

      SoMovingObj( const string& obj_name,
		   SoActiveTranslation* obj_trans,
		   const Type& a_type,
		   SoXtWebViewer* vw );

      ~SoMovingObj( void );
      const string& getName( void );
      bool visible( void );
      void show( void );
      void hide( void );
      void pause( void );
      void resume( void );
      void update( const Class& cl );
      void update( const SbVec3f& pos, const Class& cl );
      void seizeCam( SoCamera* aCam );
      void updateCam( void );
      void releaseCam( void );
      SoActiveTranslation* getPosition( void );
      void setMoveParams( const GeoPoint& beg_gp,
			  const GeoPoint& end_gp,
			  const float& time_delta );

      SoXtWebViewer* viewer_;

   private:
      static SbVec3f camDist;

      string name;
      char speed_str_[16];
      SoActiveTranslation* pos_;
      SoShape* shape;
      SoGroup* loc_;
      SoFont* font_;
      SoMaterial* material_;
      SoCamera* cam_;
      Type type;
      Class classification;
      float altitude;
      bool visible_;
};

inline SoMovingObj::
SoMovingObj( const string& obj_name,
	     SoActiveTranslation* pos,
	     const Type& a_type,
	     SoXtWebViewer* vw )
   : name( obj_name ), pos_( pos ), type( a_type ),
     visible_( false ), cam_( NULL ), shape( NULL ),
     viewer_( vw )
{
   ACE_TRACE( "(SoMovingObj::SoMovingObj)" );
   sprintf( speed_str_, " 0.0 k" );
   ACE_ASSERT( pos_ != NULL );
   addChild( pos_ );
   TRACE_NEW_2( material_, SoMaterial );
   material_->diffuseColor.setValue(SbColor(1.0, 0.0, 0.0));
   addChild(material_);
   TRACE_NEW_2( font_, SoFont );
   font_->name.setValue( "Times-Roman" );
   font_->size.setValue( 20.0 );
   addChild( font_ );
   TRACE_NEW_2( loc_, SoGroup );
   TRACE_REF( loc_ );
   SoTranslation* loc_trans;
   TRACE_NEW_2( loc_trans, SoTranslation );
   loc_trans->translation.setValue( 0.0, -2.0, 0.0 );
   loc_->addChild( loc_trans );
   SoText2* loc_text;
   TRACE_NEW_2( loc_text, SoText2 );
   loc_text->justification = SoText2::CENTER;
   loc_->addChild( loc_text );
}

inline const string& SoMovingObj::
getName( void )
{
   ACE_TRACE( "(SoMovingObj::getName)" );
   return name;
}

inline bool SoMovingObj::
visible( void )
{
   ACE_TRACE( "(SoMovingObj::visible)" );
   return visible_;
}

inline void SoMovingObj::
show( void )
{
   ACE_TRACE( "(SoMovingObj::show)" );
   if( ! visible_ )
   {
      addChild( shape );
      if( cam_ )
	 addChild( loc_ );

      visible_ = true;
   }
}

inline void SoMovingObj::
hide( void )
{
   ACE_TRACE( "(SoMovingObj::hide)" );
   if( visible_ )
   {
      if( cam_ )
	 removeChild( loc_ );

      removeChild( shape );
      visible_ = false;
   }
}

inline void SoMovingObj::
update( const SbVec3f& pos, const Class& cl )
{
   ACE_TRACE( "(SoMovingObj::update(SbVec3f, Class))" );
   update( cl );
   pos_->placeAt( pos );
   altitude = pos[1];
}

inline void SoMovingObj::
pause( void )
{
   ACE_TRACE( "(SoMovingObj::pause)" );
   pos_->pause();
}

inline void SoMovingObj::
resume( void )
{
   ACE_TRACE( "(SoMovingObj::resume)" );
   pos_->resume();
}

inline void SoMovingObj::
seizeCam( SoCamera* aCam )
{
   ACE_TRACE( "(SoMovingObj::seizeCam)" );
   ACE_ASSERT( cam_ == NULL );
   ACE_ASSERT( aCam != NULL );
   cam_ = aCam;
   cam_->position = pos_->translation.getValue() + camDist;
   cam_->pointAt( pos_->translation.getValue() );
   GeoPoint location = viewer_->getOrigin() -
      GeoPoint( pos_->translation.getValue(), viewer_->getScale() );
   ( ( SoText2* )loc_->getChild( 1 ) )->
      string.setValue( ( location.longString() + speed_str_ ).c_str() );
   addChild( loc_ );
}

inline void SoMovingObj::
releaseCam( void )
{
   ACE_TRACE( "(SoMovingObj::releaseCam)" );
   ACE_ASSERT( cam_ != NULL );
   cam_ = NULL;
   removeChild( loc_ );
}

inline SoActiveTranslation* SoMovingObj::
getPosition( void )
{
   ACE_TRACE( "(SoMovingObj::getPosition)" );
   return pos_;
}

inline void SoMovingObj::
updateCam( void )
{
   ACE_TRACE( "(SoMovingObj::updateCam)" );
   if( visible_ )
   {
      cam_->position = pos_->translation.getValue() + camDist;
      GeoPoint location = viewer_->getOrigin() -
	 GeoPoint( pos_->translation.getValue(), viewer_->getScale() );
      ( ( SoText2* )loc_->getChild( 1 ) )->
	 string.setValue( ( location.longString() + speed_str_ ).c_str() );
      viewer_->render();
   }
}

inline SoMovingObj::
~SoMovingObj( void )
{
   ACE_TRACE( "(SoMovingObj::~SoMovingObj)" );
   hide();
   removeChild( font_ );
   removeChild( pos_ );
   TRACE_UNREF( loc_ );
   if( shape )
      TRACE_UNREF( shape );

   ACE_DEBUG(( LM_DEBUG, "this=0x%x, ref=%d\n", this, getRefCount() ));
}

#endif
