/*
 * File         : SoVectors.C
 * Author       : Remek Trzaska
 * Created      : Wed Dec 17 14:34:26 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1998/02/12 20:35:44 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "SoVectors.h"
#include "iostream.h"

#define BIGNUMBER 1e10

SoVectors::SoVectors()
{};

SoVectors::SoVectors(float* aDataU, float* aDataV, float* aDataW,
		     int aSizeX, int aSizeY, int aSizeZ,
		     SoXtWebViewer* viewer,
		     int aMode,
		     float aDeltaX, float aDeltaY, float aDeltaZ) :
   SoPlaneable(aSizeX, aSizeY, aSizeZ, viewer, aMode, aDeltaX, aDeltaY, aDeltaZ)
{
   copyArray(dataU, aDataU);
   copyArray(dataV, aDataV);
   copyArray(dataW, aDataW);
   setName("vectors");
   findMinMax();
};

SoVectors::~SoVectors()
{
   delete[] dataU;
   delete[] dataV;
   delete[] dataW;
}

void
SoVectors::constructContent()
{
   int counter = 0;
   SoSeparator* sceneSep = new SoSeparator;
   addChild(sceneSep);

   translation = new SoTranslation;
   sceneSep->addChild(translation);

   coor = new SoCoordinate3;
   sceneSep->addChild(coor);

   setCoordinates();

   SoIndexedLineSet* lines = new SoIndexedLineSet;
   sceneSep->addChild(lines);
   int width;
   int height;
   
   switch (mode)
   {
      case XZ:
	 width = sizeX;
	 height = sizeZ;
	 break;
	 
      case XY:
	 width = sizeX;
	 height = sizeY;
	 break;
	 
      case YZ:
	 width = sizeZ;
	 height = sizeY;
	 break;
	 
   };
   for(int arrow = 0; arrow < height * width; arrow++)
   {
      lines->coordIndex.set1Value(counter++, arrow*2+0);
      lines->coordIndex.set1Value(counter++, arrow*2+1);
      lines->coordIndex.set1Value(counter++, -1);
   }
};

void SoVectors::
setData(float* aDataU, float* aDataV, float* aDataW,
	int aSizeX, int aSizeY, int aSizeZ)
{
   int reconstruct = (sizeX != aSizeX)||(sizeY != aSizeY)||(sizeZ != aSizeZ);
   sizeX = aSizeX; 
   sizeY = aSizeY; 
   sizeZ = aSizeZ;
   delete dataU;
   delete dataV;
   delete dataW;
   copyArray(dataU, aDataU);
   copyArray(dataV, aDataV);
   copyArray(dataW, aDataW);
   findMinMax();

   if(reconstruct)
      construct();
   else
      updateValues();

};

void
SoVectors::updateValues()
{
   setCoordinates();
};


void
SoVectors::setCoordinates()
{
   SbVec3f vec;
   int counter = 0;
   int width;
   int height;
   
   switch (mode)
   {
      case XZ:
	 width = sizeX;
	 height = sizeZ;
	 break;
	 
      case XY:
	 width = sizeX;
	 height = sizeY;
	 break;
	 
      case YZ:
	 width = sizeZ;
	 height = sizeY;
	 break;
	 
   };
   for(int i = 0; i < height; i++)
      for(int j = 0; j < width; j++)
	 switch(mode)
	 {
	    case XZ:
	       vec = arrow(j, (int)currY, i);
	       coor->point.set1Value(counter++, j*deltaX, 0, i*deltaZ);
	       coor->point.set1Value(counter++,
				     j*deltaX+vec[0], vec[1], i*deltaZ+vec[2]);
	       break;

	    case XY:
	       vec = arrow(j, i, (int)currZ);
	       coor->point.set1Value(counter++, j*deltaX, i*deltaY, 0);
	       coor->point.set1Value(counter++,
				     j*deltaX+vec[0], i*deltaY+vec[1], vec[2]);
	       break;
	       
	    case YZ:
	       vec = arrow((int)currX, i, j);
	       coor->point.set1Value(counter++, 0, i*deltaY, j*deltaZ);
	       coor->point.set1Value(counter++,
				     vec[0], i*deltaY+vec[1], j*deltaZ+vec[2]);
	       break;
	       
	 }
};

void
SoVectors::findMinMax()
{
   maxLength = 0.0;
   
   for(int i = 0; i < sizeX * sizeY * sizeZ; i++)
   {
      float valueU = dataU[i];
      float valueV = dataV[i];
      float valueW = dataW[i];
      
      if(valueU >  BIGNUMBER) dataU[i] = 0.0;
      if(valueU < -BIGNUMBER) dataU[i] = 0.0;
      
      if(valueV >  BIGNUMBER) dataV[i] = 0.0;
      if(valueV < -BIGNUMBER) dataV[i] = 0.0;

      if(valueW >  BIGNUMBER) dataW[i] = 0.0;
      if(valueW < -BIGNUMBER) dataW[i] = 0.0;

      if(maxLength==0.0)
	 maxLength = sqrt(valueW*valueW + valueV*valueV + valueU*valueU);
      
   }
};


SbVec3f
SoVectors::arrow(int x, int y, int z)
{
   if(maxLength == 0) return SbVec3f(0,0,0);
   
   float xv = dataU[y*sizeZ*sizeX + x*sizeZ + z];
   float yv = dataV[y*sizeZ*sizeX + x*sizeZ + z];
   float zv = dataW[y*sizeZ*sizeX + x*sizeZ + z];

   xv = xv * deltaX / 30.0;
   yv = yv * deltaY / 30.0;
   zv = zv * deltaZ / 30.0;

   return SbVec3f(xv, yv, zv);
};


void
SoVectors::copyArray(float* &dest, float* source)
{
   dest = new float[sizeX * sizeY * sizeZ];
   memcpy(dest, source, sizeX * sizeY * sizeZ * sizeof(float));
};

