/* $Id: TagParser.C,v 1.31 1998/04/23 20:13:40 remek Exp $ */
/*
 * File:       TagParser.C
 * Author:     Pawel Roman (proman@npac.syr.edu),
 *             Remek Trzaska <remek@npac.syr.edu>
 * Created:    long time ago
 * Revision:   $Revision: 1.31 $
 * Updated:    $Date: 1998/04/23 20:13:40 $
 * Copyright 1996/97 NPAC, Syracuse University
 *
 */

#include "TagParser.h"
#include "GeoPoint.h"
#include "SoChunk.h"
#include "SoXtWebViewer.h"
#include "RectArea.h"
#include "CircleArea.h"

#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoInfo.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>

#include <STL/stlstring.h>

int
rootTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(rootTagCB)" );
   int result = 0;
   if( list.getLength() > 1 )
   {
      ACE_ERROR(( LM_ERROR, "[!](rootTagCB) Found %d GIS_Root tags\n",
		  list.getLength() ));
      result = -1;
   }
   SoInfo* a_info = ( SoInfo* )list[0];
   ACE_DEBUG(( LM_DEBUG, "Root: %s\n", a_info->string.getValue().getString() ));
   viewer.setGISRoot( a_info->string.getValue().getString() );
   for( int i=0 ; i < list.getLength(); i++ )
      viewer.getGraphRoot()->removeChild( a_info );
   
   return result;
}


int
scaleTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(scaleTagCB)" );
   int result = 0;
   double lon_deg_size, lat_deg_size, alt_meter_size;
   
   if( list.getLength() > 1 )
   {
      ACE_ERROR(( LM_ERROR, "[!](scaleTagCB) Found %d GIS_Scale tags\n",
		  list.getLength() ));
      result = -1;
   }
   SoInfo* a_info = ( SoInfo* )list[0];
   sscanf( a_info->string.getValue().getString(), "%lf %lf %lf",
	   &lon_deg_size, &lat_deg_size, &alt_meter_size );

   ACE_DEBUG(( LM_DEBUG, "Scale: (%.2f, %.2f, %.2f)\n",
	       lon_deg_size, lat_deg_size, alt_meter_size ));

   viewer.setScale( CoordScale( lon_deg_size, lat_deg_size, alt_meter_size ) );
   for( int i=0 ; i < list.getLength(); i++ )
      viewer.getGraphRoot()->removeChild( a_info );
   
   return result;
}


int
originTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(originTagCB)" );
   int result = 0;
   int lon_deg, lon_min, lon_sec,
      lat_deg, lat_min, lat_sec;
   double alt_meter;
   if( list.getLength() > 1 )
   {
      ACE_ERROR(( LM_ERROR, "[!](originTagCB) Found %d GIS_Origin tags\n",
		  list.getLength() ));
      result = -1;
   }
   SoInfo* a_info = ( SoInfo* )list[0];
   sscanf( a_info->string.getValue().getString(), "%d:%d:%d,%d:%d:%d %lf",
	   &lon_deg, &lon_min, &lon_sec,
	   &lat_deg, &lat_min, &lat_sec,
	   &alt_meter );

   ACE_DEBUG(( LM_DEBUG, "Origin: (%d:%d:%d %d:%d:%d %.2f)\n",
	       lon_deg, lon_min, lon_sec,
	       lat_deg, lat_min, lat_sec,
	       alt_meter ));

   viewer.setOrigin( GeoPoint( lon_deg, lon_min, lon_sec,
			       lat_deg, lat_min, lat_sec,
			       alt_meter ) );
   for( int i=0 ; i < list.getLength(); i++ )
      viewer.getGraphRoot()->removeChild( a_info );
   
   return result;
}

//Important!!! GIS_Bounds tag has to appear before GIS_Terrain
//in order to function properly

int
boundsTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(boundsTagCB)" );
   int result = 0;
   int lon_deg1, lon_min1, lon_sec1, lat_deg1, lat_min1, lat_sec1;
   int lon_deg2, lon_min2, lon_sec2, lat_deg2, lat_min2, lat_sec2;

   if( list.getLength() > 1 )
   {
      ACE_ERROR(( LM_ERROR, "[!](boundsTagCB) Found %d GIS_Bounds tags\n",
		  list.getLength() ));
      result = -1;
   }
   SoInfo* a_info = ( SoInfo* )list[0];
   sscanf( a_info->string.getValue().getString(),
	   "%d:%d:%d,%d:%d:%d %d:%d:%d,%d:%d:%d ",
	   &lon_deg1, &lon_min1, &lon_sec1, &lat_deg1, &lat_min1, &lat_sec1,
	   &lon_deg2, &lon_min2, &lon_sec2, &lat_deg2, &lat_min2, &lat_sec2);

   ACE_DEBUG(( LM_DEBUG, "Bounds: (%d:%d:%d %d:%d:%d %d:%d:%d %d:%d:%d )\n",
	       lon_deg1, lon_min1, lon_sec1, lat_deg1, lat_min1, lat_sec1,
	       lon_deg2, lon_min2, lon_sec2, lat_deg2, lat_min2, lat_sec2));
   

   viewer.setBounds(
      GeoPoint(lon_deg1,lon_min1,lon_sec1,lat_deg1,lat_min1,lat_sec1,0),
      GeoPoint(lon_deg2,lon_min2,lon_sec2,lat_deg2,lat_min2,lat_sec2,0));

   viewer.getGraphRoot()->removeChild( a_info );
   
   return result;
}


int
textureTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(textureTagCB)" );
   int result = 0;

   vector<string>* lref;
   vector<string>* ldesc;
   int nCount = list.getLength();
   TRACE_NEW_2( lref, vector<string>( nCount ) );
   TRACE_NEW_2( ldesc, vector<string>( nCount ) );
   for( int i=0; i < nCount; i++ )
   {
      SoInfo* a_info = ( SoInfo* )list[i];
      const char* texInfo = a_info->string.getValue().getString();
      const char* sep = strchr( texInfo, TOKEN_SEPARATOR );
      ( *lref )[i] = string( texInfo, sep-texInfo );
      ( *ldesc )[i] = string( sep+1 );
      ACE_DEBUG(( LM_DEBUG, "Texture: %s as %s\n", ( *ldesc )[i].c_str(),
		  ( *lref )[i].c_str() ));
      viewer.getGraphRoot()->removeChild( a_info );
   }
   viewer.makeLayerList( lref, ldesc );
   return result;
}


int
rectAreaTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(rectAreaTagCB)" );
   int result = 0;
   int lon_deg_UL, lon_min_UL, lon_sec_UL,
      lat_deg_UL, lat_min_UL, lat_sec_UL,
      lon_deg_BR, lon_min_BR, lon_sec_BR,
      lat_deg_BR, lat_min_BR, lat_sec_BR;
   int nCount = list.getLength();
   vector<Area*> rects( nCount );
   for( int i=0 ; i < nCount; i++ )
   {
      SoInfo* a_info = ( SoInfo* )list[i];
      sscanf( a_info->string.getValue().getString(), "%d:%d:%d,%d:%d:%d %d:%d:%d,%d:%d:%d",
	      &lon_deg_UL, &lon_min_UL, &lon_sec_UL,
	      &lat_deg_UL, &lat_min_UL, &lat_sec_UL,
	      &lon_deg_BR, &lon_min_BR, &lon_sec_BR,
	      &lat_deg_BR, &lat_min_BR, &lat_sec_BR );

      const char* ref_str = strchr( a_info->string.getValue().getString(), ',' )+1;
      if( ! ref_str )
      {
	 ACE_ERROR(( LM_ERROR, "[!](TagParser::rectAreaTagCB) Cannot retrieve ref from: %s\n",
		     a_info->string.getValue().getString() ));
	 viewer.getGraphRoot()->removeChild( a_info ); 
	 continue;
      }
      string ref( ref_str );
      
      ACE_DEBUG(( LM_DEBUG, "RectArea: [%d:%d:%d %d:%d:%d] [%d:%d:%d %d:%d:%d] ref: %s\n",
		  lon_deg_UL, lon_min_UL, lon_sec_UL,
		  lat_deg_UL, lat_min_UL, lat_sec_UL,
		  lon_deg_BR, lon_min_BR, lon_sec_BR,
		  lat_deg_BR, lat_min_BR, lat_sec_BR,
		  ref.c_str() ));

      TRACE_NEW_2( rects[i],
		   RectArea( GeoPoint( lon_deg_UL, lon_min_UL, lon_sec_UL,
				       lat_deg_UL, lat_min_UL, lat_sec_UL, 0 ),
			     GeoPoint( lon_deg_BR, lon_min_BR, lon_sec_BR,
				       lat_deg_BR, lat_min_BR, lat_sec_BR, 0 ),
			     ref ) );

      viewer.getGraphRoot()->removeChild( a_info ); 
   }
   viewer.addAreas( rects );
   return result;
}


int
circleAreaTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(circleAreaTagCB)" );
   int result = 0;
   int lon_deg_CTR, lon_min_CTR, lon_sec_CTR,
      lat_deg_CTR, lat_min_CTR, lat_sec_CTR,
      lon_deg_MRK, lon_min_MRK, lon_sec_MRK,
      lat_deg_MRK, lat_min_MRK, lat_sec_MRK;
   int nCount = list.getLength();
   vector<Area*> circles( nCount );
   for( int i=0 ; i < nCount; i++ )
   {
      SoInfo* a_info = ( SoInfo* )list[i];
      sscanf( a_info->string.getValue().getString(), "%d:%d:%d,%d:%d:%d %d:%d:%d,%d:%d:%d",
	      &lon_deg_CTR, &lon_min_CTR, &lon_sec_CTR,
	      &lat_deg_CTR, &lat_min_CTR, &lat_sec_CTR,
	      &lon_deg_MRK, &lon_min_MRK, &lon_sec_MRK,
	      &lat_deg_MRK, &lat_min_MRK, &lat_sec_MRK );

      const char* ref_str = strchr( a_info->string.getValue().getString(), ',' )+1;
      if( ! ref_str )
      {
	 ACE_ERROR(( LM_ERROR, "[!](TagParser::circleAreaTagCB) Cannot retrieve ref: %s\n",
		     a_info->string.getValue().getString() ));
	 viewer.getGraphRoot()->removeChild( a_info ); 
	 continue;
      }
      string ref( ref_str );

      ACE_DEBUG(( LM_DEBUG, "CircleArea: [%d:%d:%d %d:%d:%d] [%d:%d:%d %d:%d:%d] ref: %s\n",
		  lon_deg_CTR, lon_min_CTR, lon_sec_CTR,
		  lat_deg_CTR, lat_min_CTR, lat_sec_CTR,
		  lon_deg_MRK, lon_min_MRK, lon_sec_MRK,
		  lat_deg_MRK, lat_min_MRK, lat_sec_MRK,
		  ref.c_str() ));

      TRACE_NEW_2( circles[i],
		   CircleArea( GeoPoint( lon_deg_CTR, lon_min_CTR, lon_sec_CTR,
					 lat_deg_CTR, lat_min_CTR, lat_sec_CTR, 0 ),
			       GeoPoint( lon_deg_MRK, lon_min_MRK, lon_sec_MRK,
					 lat_deg_MRK, lat_min_MRK, lat_sec_MRK, 0 ),
			       ref ) );

      viewer.getGraphRoot()->removeChild( a_info ); 
   }
   viewer.addAreas( circles );
   return result;
}


int
objectTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(objectTagCB)" );
   int result = 0;
   int lon_d, lon_m, lon_s, lat_d, lat_m, lat_s;

   int nCount = list.getLength();
   
   for( int i=0; i < nCount; i++ )
   {
      SoSeparator* objectGroup = ( SoSeparator* )list[i];
      //check format
      if(!objectGroup->isOfType(SoSeparator::getClassTypeId()))
      {
	 ACE_ERROR(( LM_ERROR, "[!] objectTag not being SoSeparator: %s\n",
		     objectGroup->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(objectGroup);
	 break;
      }
      if(objectGroup->getNumChildren()!=2 && objectGroup->getNumChildren()!=3 )
      {
	 ACE_ERROR(( LM_ERROR, "[!] objectTag not having exactly 2 or 3 children: %s\n",
		     objectGroup->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(objectGroup);
	 break;
      }
      if(!objectGroup->getChild(0)->isOfType(SoInfo::getClassTypeId()))
      {
	 ACE_ERROR(( LM_ERROR, "[!] objectTag 1st child is not an Info node: %s\n",
		     objectGroup->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(objectGroup);
	 break;
      }
      SoInfo* info = (SoInfo*)objectGroup->getChild(0);
      sscanf(info->string.getValue().getString(), "%d:%d:%d,%d:%d:%d",
	     &lon_d, &lon_m, &lon_s,
	     &lat_d, &lat_m, &lat_s);
      
      GeoPoint objectsOrigin(lon_d, lon_m, lon_s, lat_d, lat_m, lat_s, 0);

      //translate geocoodrs to translation
      GeoPoint gp = viewer.getOrigin() - objectsOrigin;
      
      SoTranslation* translation = new SoTranslation;
      translation->translation.setValue(gp.getScaledVec( viewer.getScale()));
      SbVec3f transVec = translation->translation.getValue();
      if( viewer.pointOnSurface( transVec ) )
      {
	 translation->translation.setValue(transVec);
	 objectGroup->replaceChild(info, translation);
      }
      else
      {
	 translation->translation.setValue(transVec);
	 objectGroup->replaceChild(info, translation);
      }
   }
   return result;
};


int
terrainTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(terrainTagCB)" );
   int result = 0;
   int lon_deg_UL, lon_min_UL, lon_sec_UL,
      lat_deg_UL, lat_min_UL, lat_sec_UL,
      demSize, texSize;
   float quality;
   float yScale = viewer.getScale().alt_meter_size;
   int nCount = list.getLength();
   vector<SoChunk*>* chunks;
   TRACE_NEW_2( chunks, vector<SoChunk*>( nCount ) );
   for( int i=0 ; i < nCount; i++ )
   {
      SoInfo* a_info = ( SoInfo* )list[i];
      sscanf( a_info->string.getValue().getString(), "%d:%d:%d,%d:%d:%d %d %f %d",
	      &lon_deg_UL, &lon_min_UL, &lon_sec_UL,
	      &lat_deg_UL, &lat_min_UL, &lat_sec_UL,
	      &demSize, &quality, &texSize );
      ACE_DEBUG(( LM_DEBUG, "Terrain: (%d:%d:%d %d:%d:%d) %d %f %d\n",
		  lon_deg_UL, lon_min_UL, lon_sec_UL,
		  lat_deg_UL, lat_min_UL, lat_sec_UL,
		  demSize, quality, texSize ));
      GeoPoint corner( lon_deg_UL, lon_min_UL, lon_sec_UL,
		       lat_deg_UL, lat_min_UL, lat_sec_UL, 0 );

      float maxWest = viewer.getMaxWest();
      float maxNorth = viewer.getMaxNorth();
      float maxEast = viewer.getMaxEast();
      float maxSouth = viewer.getMaxSouth();
      
      TRACE_NEW_2( ( *chunks )[i],
		   SoChunk( corner,
 			    demSize,
			    quality,
			    texSize,
			    yScale,
			    viewer.getBoundsTagParsed(),
			    maxWest, maxNorth, maxEast, maxSouth));

      viewer.recalculateBounds(corner, demSize);
      viewer.getGraphRoot()->removeChild( a_info );
   }
   viewer.setChunks( chunks );
   return result;
}


// sliderTag has to be a separator with:
//   info with a vector of the axis
//   translation to starting position
//   geometry group
// for example:
//   DEF GIS_sliderTag Separator {
//     Info {
//       string "1.0 1.0 0"
//     }
//     Transform {
//       rotation 1.0 0.2 0.3 3.14
//     }
//     Group {
//       Cone {
//       }
//     }
//   }

int
sliderTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(sliderTagCB)" );
   float dirX, dirY, dirZ;

   int nCount = list.getLength();
   
   for( int i=0; i < nCount; i++ )
   {
      SoSeparator* sliderSep = ( SoSeparator* )list[i];
      //check format
      if(!sliderSep->isOfType(SoSeparator::getClassTypeId()))
      {
	 ACE_ERROR(( LM_ERROR, "[!] sliderTag not being SoSeparator: %s\n",
		     sliderSep->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(sliderSep);
	 break;
      }
      if(sliderSep->getNumChildren()!=3)
      {
	 ACE_ERROR(( LM_ERROR, "[!] sliderTag not having exactly 3 children: %s\n",
		     sliderSep->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(sliderSep);
	 break;
      }
      if(!sliderSep->getChild(0)->isOfType(SoInfo::getClassTypeId()))
      {
	 ACE_ERROR(( LM_ERROR, "[!] sliderTag 1st child is not an Info node: %s\n",
		     sliderSep->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(sliderSep);
	 break;
      }
      if(!sliderSep->getChild(1)->isOfType(SoTransform::getClassTypeId()))
      {
	 ACE_ERROR(( LM_ERROR, "[!] sliderTag 2nd child is not an Transform node: %s\n",
		     sliderSep->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(sliderSep);
	 break;
      }
      if(!sliderSep->getChild(2)->isOfType(SoGroup::getClassTypeId()))
      {
	 ACE_ERROR(( LM_ERROR, "[!] sliderTag 3rd child is not a Group node: %s\n",
		     sliderSep->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(sliderSep);
	 break;
      }
      SoInfo* info = (SoInfo*)sliderSep->getChild(0);
      SoGroup* geometry = (SoGroup*)sliderSep->getChild(2);

      //find out what the direction is
      sscanf(info->string.getValue().getString(), "%f %f %f",
	     &dirX, &dirY, &dirZ);

      if(dirX==0.0 && dirY==0.0 && dirZ==0.0)
      {
	 ACE_ERROR(( LM_ERROR, "[!] direction vector is NULL: %s\n",
		     sliderSep->getName().getString() ));
	 viewer.getGraphRoot()->removeChild(sliderSep);
	 break;
      }

      //add the rotation
      SoRotation* rotation = new SoRotation;
      
      rotation->rotation.setValue(SbRotation(SbVec3f(1, 0, 0),
					     SbVec3f(dirX, dirY, dirZ)));
      
      
      sliderSep->insertChild(rotation, 2);

      //add magnified dragger
      SoSeparator* boxTranslator = new SoSeparator;
      TRACE_REF(boxTranslator);
      boxTranslator->addChild(geometry);
      
      SoSeparator* activeBoxTranslator = new SoSeparator;
      TRACE_REF(activeBoxTranslator);
      SoMaterial *activeBoxMaterial = new SoMaterial;
      activeBoxMaterial->diffuseColor.setValue(1, 1, 0);
      activeBoxTranslator->addChild(activeBoxMaterial);   
      activeBoxTranslator->addChild(geometry);
      
      SoTranslate1Dragger* dragger = new SoTranslate1Dragger;
      sliderSep->insertChild(dragger, 3);
      
      //add transform
      SoTransform* transform = new SoTransform;
      sliderSep->insertChild(transform, 4);
      transform->translation.connectFrom(&dragger->translation);

      dragger->setPart("translator", boxTranslator);
      dragger->setPart("translatorActive", activeBoxTranslator);

      TRACE_UNREF(boxTranslator);
      TRACE_UNREF(activeBoxTranslator);

      SoRotation* antiRotation = new SoRotation;
      SbRotation sbRot = rotation->rotation.getValue();
      
      antiRotation->rotation.setValue(sbRot.invert());
      
      geometry->insertChild(antiRotation, 0);
      sliderSep->removeChild(geometry);

      dragger->addMotionCallback( ( SoDraggerCB& )SoXtWebViewer::draggerMovedCB,
				  &viewer );
      char buf[32];
      sprintf(buf, "dragger_%d", i);
      dragger->setName(SbName(buf));
   }
   return 1;
};


#ifdef V5DSTUFF

int
v5dFileTagCB( SoNodeList& list, SoXtWebViewer& viewer )
{
   ACE_TRACE( "(v5dFileTagCB)" );
   int result = 0;
   if( list.getLength() > 1 )
   {
      ACE_ERROR(( LM_ERROR, "[!](v5dFileTagCB) Found %d GIS_V5DFile tags\n",
		  list.getLength() ));
      result = -1;
   }
   SoInfo* a_info = ( SoInfo* )list[0];
   ACE_DEBUG(( LM_DEBUG, "V5DFile: %s\n",
	       a_info->string.getValue().getString() ));
   
   char v5dFileName[512];
   strncpy(v5dFileName, a_info->string.getValue().getString(), 511);
   viewer.setV5DFile( v5dFileName );
   viewer.getGraphRoot()->removeChild( a_info );
   return result;
}

#endif //V5DSTUFF


const TagAction TagParser::tags[] = {
   { "GIS_Root", rootTagCB },
   { "GIS_Origin", originTagCB },
   { "GIS_Bounds", boundsTagCB },
   { "GIS_Scale", scaleTagCB },
   { "GIS_Terrain", terrainTagCB },
   { "GIS_Texture", textureTagCB },
   { "GIS_RectArea", rectAreaTagCB },
   { "GIS_CircleArea", circleAreaTagCB },
   { "GIS_Object", objectTagCB },
   { "GIS_Slider", sliderTagCB }
#ifdef V5DSTUFF   
   ,
   { "GIS_V5DFile", v5dFileTagCB }
#endif //V5DSTUFF
};

enum TAG_RESULT { ALL_OK, TAG_NOT_FOUND };

int TagParser::
parse( void )
{
   ACE_TRACE( "(TagParser::parse)" );
   int result = ALL_OK;
   ACE_DEBUG(( LM_DEBUG, "%d tags registered\n", sizeof tags / sizeof tags[0] ));
   SoNodeList* ptagList;
   TRACE_NEW_2( ptagList, SoNodeList );
   SoNodeList& tagList = *ptagList;
   viewer.setAutoRedraw( FALSE );
   for( int i=0; i < sizeof tags / sizeof tags[0]; i++ )
   {
      ACE_DEBUG(( LM_DEBUG, "before retreiving %s tag node(s)\n", tags[i].tag ));
      SoNode::getByName( tags[i].tag, *ptagList );
      ACE_DEBUG(( LM_DEBUG, "after making tagList\n" ));
      if( tagList.getLength() == 0 )
      {
	 ACE_ERROR(( LM_ERROR, "[!](TagParser::parse) %s tag not found\n", tags[i].tag ));
	 result = TAG_NOT_FOUND;
      }
      else
      {
	 ACE_DEBUG(( LM_DEBUG, "tags[0] ref before action: %d\n", tagList[0]->getRefCount() ));
	 tags[i].action( tagList, viewer );
	 ACE_DEBUG(( LM_DEBUG, "tags[0] ref after action: %d\n", tagList[0]->getRefCount() ));
      }
      tagList.truncate( 0 );
   }
   TRACE_DELETE( ptagList );
   viewer.render();
   viewer.setAutoRedraw( TRUE );
   return result;
}

