/* $Id: WebNode.C,v 1.22 1997/10/01 19:35:46 proman Exp $ */
/*
 * File:       WebNode.C
 * Author:     Pawel Roman (proman@npac.syr.edu)
 * Created:    long time ago
 * Revision:   $Revision: 1.22 $
 * Updated:    $Date: 1997/10/01 19:35:46 $
 * Copyright 1996/97 NPAC, Syracuse University
 *
 */

#include "WebNode.h"

#include <wv.h>
#include <Inventor/Xt/SoXt.h>

static const char OK_CODE[] = "200";       // status code for OK in HTTP/1.x
static const char MPERM_CODE[] = "301";    // status code for Moved Perm in HTTP/1.x
static const char MTEMP_CODE[] = "302";    // status code for Moved Temp in HTTP/1.x
static const char GET_COMM[] = "GET ";
static const char HTTP_SUFFIX[] = " HTTP/1.0\n\n";
static const char LOC_FIELD[] = "Location:";
static const char CONT_FIELD_1[] = "Content-length:";
static const char CONT_FIELD_2[] = "Content-Length:";
static const char CAR_RETURN[] = "\r";

enum {
   WN_OK,
   WN_EINVURL,
   WN_ECONN,
   WN_EDLOAD,
   WN_EHTTP,
   WN_ENOMEM,
   WN_UADC
};

static const char* WN_ERROR[] = {
   "No error",
   "Invalid URL",
   "Unable to connect",
   "Downloading failure",
   "HTTP Protocol failure: No 'Content-length' field",
   "No memory for allocation",
   "Unable to uncompress"
};

static const int defaultDEMQuality = 9;

Cache WebNode::cache_;

WebNode::
WebNode( const char* location, CallbackFunc cb, void* ud, int aRotate ) 
   : server_( NULL ), URL_( NULL ), dataBuf_( NULL ), dataSize_( 0 ), fetched_( 0 ),
     doneCB_( cb ), userData_( ud ), bufGiven_( false ), intStatus_( WN_OK ),
     hdRecved_( false ), rotate( aRotate )
{ 
   ACE_TRACE( "(WebNode::WebNode)" );
   if( ! cache_.isCached( location ) )
   {
      TRACE_NEW_2( URL_, URL( location ) );
      if( ! URL_ )
      {
	 ACE_ERROR(( LM_ERROR, "[!](WebNode::WebNode) %s: URL_\n", WN_ERROR[WN_ENOMEM] ));
	 intStatus_ = WN_ENOMEM;
	 if( doneCB_ )
	    doneCB_( this, userData_ );
      }
      else if( ! *URL_ )
      {
	 strncpy( buf_, "", MAXLEN );
	 ACE_ERROR(( LM_ERROR, "[!](WebNode::WebNode) %s\n", WN_ERROR[WN_EINVURL] ));
	 intStatus_ = WN_EINVURL;
	 if( doneCB_ )
	    doneCB_( this, userData_ );
      }
      else
	 download();
   }
   else
   {
      dataBuf_ = cache_.getBuffer();
      dataSize_ = cache_.getBufferSize();
      intStatus_ = WN_OK;
      if( doneCB_ )
	 doneCB_( this, userData_ );
   }
}

WebNode::
~WebNode( void )
{
   ACE_TRACE( "(WebNode::~WebNode)" );
   ACE_DEBUG(( LM_DEBUG, "this=0x%x\n", this ));
   if( server_ )
   {
      server_->close();
      TRACE_DELETE( server_ );
   }
   if( URL_ )
      TRACE_DELETE( URL_ );

   if( ! bufGiven_ )
      TRACE_DELETE( dataBuf_ );
}

const char* WebNode::
getErrMsg( void ) const
{
   return WN_ERROR[intStatus_];
}

void WebNode::
finish( void )
{
   ACE_TRACE( "(WebNode::finish)" );
   //if compressed (extension ends with wv), uncompress it
   const char* location = SbString( *URL_ ).getString();
   const char* extension = strrchr( location, '.' );
   if( extension!=NULL )
   {
      if( extension[1] == 't' &&extension[2]=='w' && extension[3]=='v')
	 uncompressTex();
      if( extension[1] == 'd' &&extension[2]=='w' && extension[3]=='v')
	 uncompressDEM();
      if( extension[1] == 'p' &&extension[2]=='p' && extension[3]=='m')
	 processPPM();  //convert it to the raw form
#ifdef TRANSFORM_TEXTURES
      if((extension[1] == 'p' &&extension[2]=='p' && extension[3]=='m') ||
	 (extension[1] == 't' &&extension[2]=='w' && extension[3]=='v'))
	 transformTex();
#endif //TRANSFORM_TEXTURES
   }
   cache_.cacheIt( location, dataBuf_, dataSize_ );
}

void WebNode::
download( void )
{
   ACE_TRACE( "(WebNode::download)" );
   TRACE_NEW_2( server_, ACE_SOCK_IOStream );
   if( ! server_ )
   {
      ACE_ERROR(( LM_ERROR, "[!](WebNode::download) No memory for server_\n" ));
      intStatus_ = WN_ENOMEM;
      return;
   }

   ACE_SOCK_Connector connector;
   ACE_INET_Addr addr( URL_->getPort(), URL_->getHost().getString() );

   if( connector.connect( *server_, addr ) == -1 )
   {
      strncpy( buf_, strerror( errno ), MAXLEN );
      ACE_ERROR(( LM_ERROR, "[!](WebNode::download) %s: %s\n", WN_ERROR[WN_ECONN], buf_ ));
      intStatus_ = WN_ECONN;
      return;
   }

   ACE_SOCK_IOStream& sock = *server_;
   sock << GET_COMM << URL_->getResource().getString() << HTTP_SUFFIX;
   sock.flush();
   if( doneCB_ )
      XtAppAddInput( SoXt::getAppContext(),
		     server_->get_handle(),
		     ( XtPointer )XtInputReadMask,
		     ( XtInputCallbackProc )asynch_input,
		     ( XtPointer )this );
   else
      synch_input();
}

int WebNode::
recvHeader( void )
{
   ACE_TRACE( "(WebNode::recvHeader)" );
   ACE_SOCK_IOStream& sock = *server_;
   sock.getline( buf_, MAXLEN );
   if( strstr( buf_, MPERM_CODE ) || strstr( buf_, MTEMP_CODE ) )
   {
      while( sock.getline( buf_, MAXLEN, '\r' ) )
	 if( strstr( buf_, LOC_FIELD ) )
	 {
	    ACE_DEBUG(( LM_NOTICE, "(WebNode::recvHeader) Resource moved, %s\n", buf_+1 ));
	    if( doneCB_ )
	       ACE_ERROR_RETURN(( LM_ERROR, "[!](WebNode::recvHeader) Asynchronous mode\n",
				  WN_ERROR[WN_EDLOAD] ), WN_EDLOAD );
	    else
	    {
	       TRACE_DELETE( server_ );
	       TRACE_DELETE( URL_ );
	       TRACE_NEW_2( URL_, URL( strchr( buf_, ':' )+2 ) );
	       
	       if( ! URL_ )
		  ACE_ERROR_RETURN(( LM_ERROR, "[!](WebNode::recvHeader) %s: URL_\n",
				     WN_ERROR[WN_ENOMEM] ),
				   WN_ENOMEM );
	       
	       if( ! *URL_ )
	       {
		  strncpy( buf_, SbString( *URL_ ).getString(), MAXLEN );
		  ACE_ERROR_RETURN(( LM_ERROR, "[!](WebNode::recvHeader) %s\n",
				     WN_ERROR[WN_EINVURL] ),
				   WN_EINVURL );
	       }
	       download();
	    }
	 }
   }
   if( !strstr( buf_, OK_CODE ) )
      ACE_ERROR_RETURN(( LM_ERROR, "[!](WebNode::recvHeader) %s: %s\n",
			 WN_ERROR[WN_EDLOAD], buf_ ),
		       WN_EDLOAD );

   dataSize_ = -1;
   ACE_DEBUG(( LM_DEBUG, "(WebNode::recvHeader) HTTP header:\n" ));
   while( sock.getline( buf_, MAXLEN ) && !( ( buf_[0] == '\r' )||( buf_[0] == '\0' ) ) )
   {
      ACE_DEBUG(( LM_DEBUG, "\t%s\n", buf_ ));
      if( strstr( buf_, CONT_FIELD_1 ) || strstr( buf_, CONT_FIELD_2 ) )
      {
	 dataSize_ = atoi( strchr( buf_, ':' )+1 );
	 break;
      }
   }
   if( dataSize_ == -1 )
      ACE_ERROR_RETURN(( LM_ERROR, "[!](WebNode::recvHeader) %s\n", WN_ERROR[WN_EHTTP] ),
		       WN_EHTTP );
   
   TRACE_NEW_2( dataBuf_, char[dataSize_] );
   if( ! dataBuf_ )
      ACE_ERROR_RETURN(( LM_ERROR, "[!](WebNode::recvHeader) %s\n", WN_ERROR[WN_ENOMEM] ),
		       WN_ENOMEM );

   //discarding rest of HTTP header
   while( sock.getline( buf_, MAXLEN ) && !( ( buf_[0] == '\r' )||( buf_[0] == '\0' ) ) )
//      ACE_DEBUG(( LM_DEBUG, "\t%s\n", buf_ ));
      ACE_HEX_DUMP(( LM_DEBUG, buf_, sizeof buf_ ));
   

   hdRecved_ = true;
   return WN_OK;
}

void WebNode::
asynch_input( WebNode* wn, int*, XtInputId* id )
{
   ACE_TRACE( "(WebNode::asynch_input)" );
   ACE_DEBUG(( LM_DEBUG, "URL: %s\n", SbString( *wn->URL_ ).getString() ));
   if( ! wn->hdRecved_ )
   {
      if( ( wn->intStatus_ = wn->recvHeader() ) == WN_OK )
      {
	 if( ( wn->fetched_ = wn->syncStreamBuf() ) == wn->dataSize_ )
	 {
	    wn->finish();
	    wn->cleanUp( *id );
	 }
      }
      else
	 wn->cleanUp( *id );
   }
   else
   {
      ssize_t st = ( ( ACE_SOCK_Stream* )wn->server_ )->recv( wn->dataBuf_ + wn->fetched_,
							      wn->dataSize_ - wn->fetched_ );
      if( st > 0 )
      {
	 wn->fetched_ += st;
	 ACE_DEBUG(( LM_DEBUG, "%d byte(s) read, %d byte(s) to read\n",
		     st, wn->dataSize_ - wn->fetched_ ));
	 
	 if( wn->fetched_ == wn->dataSize_ )
	 {
	    wn->finish();
	    wn->cleanUp( *id );
	 }
      }
      else
      {
	 strncpy( wn->buf_, SbString( *wn->URL_ ).getString(), MAXLEN );
	 ACE_ERROR(( LM_ERROR, "[!](WebNode::asynch_input) %s: %s\n",
		     WN_ERROR[WN_EDLOAD], wn->buf_ ));
	 wn->intStatus_ = WN_EDLOAD;
	 wn->cleanUp( *id );
      }
   }
}

void WebNode::
synch_input( void )
{
   ACE_TRACE( "(WebNode::synch_input)" );
   if( ( intStatus_ = recvHeader() ) == WN_OK )
   {
      if( ( fetched_ = syncStreamBuf() ) < dataSize_ )
	 if( ( ( ACE_SOCK_Stream* )server_ )->recv_n( dataBuf_ + fetched_,
						      dataSize_ - fetched_ ) <= 0 )
	 {
	    strncpy( buf_, SbString( *URL_ ).getString(), MAXLEN );
	    ACE_ERROR(( LM_ERROR, "[!](WebNode::synch_input) %s: %s\n",
			WN_ERROR[WN_EDLOAD], buf_ ));
	    intStatus_ = WN_EDLOAD;
	    return;
	 }
      finish();
   }
}

void WebNode::
uncompressTex()
{
   ACE_TRACE( "(WebNode::uncompressTex)" );
   //texture header: 0,1 - dimy, 2,3 - dimx
   int dimy = (int)dataBuf_[0] * 256 + dataBuf_[1];
   int dimx = (int)dataBuf_[2] * 256 + dataBuf_[3];
   
   char* uncompressedData;
   TRACE_NEW_2( uncompressedData, char[dimy*dimx*3+2] );
   char* rgbData = uncompressedData + 2;
   uncompressedData[0] = dimy/256;
   uncompressedData[1] = dimy%256;
   
   twv_t* compressor = new_twv(dimy, dimx);

   int returnedQuality = defaultDEMQuality;
   int scale = 1;
   
   returnedQuality = dec_twv(compressor,
			     (unsigned char*)dataBuf_ + 4,
			     &dataSize_,
			     (unsigned char*)rgbData,
			     returnedQuality,
			     scale);

   if(returnedQuality<0)
   {
      intStatus_ = WN_UADC;
      memset(rgbData, dimx*dimy*3, 128);
   }
   
   delete_twv(&compressor);

   TRACE_DELETE_TAB( dataBuf_ );
   dataSize_ = dimx*dimy*3+2;
   dataBuf_ = uncompressedData;
}

#ifdef TRANSFORM_TEXTURES
void WebNode::
transformTex()
{
   int dim = dataBuf_[0]*256+dataBuf_[1];
   char* mirrTex;
   TRACE_NEW_2( mirrTex, char[dim*dim*3+2] );
   mirrTex[0] = dataBuf_[0];
   mirrTex[1] = dataBuf_[1];
   for(int row=0; row<dim; row++)
      for(int col=0; col<dim; col++)
      {
	 int newrow = col;
	 int newcol = row;
	 if(rotate)
	 {
	    newrow = row;
	    newcol = col;
	 }
	 mirrTex[(newrow*dim+newcol)*3+2] = dataBuf_[(row*dim+col)*3+2];
	 mirrTex[(newrow*dim+newcol)*3+2+1] = dataBuf_[(row*dim+col)*3+2+1];
	 mirrTex[(newrow*dim+newcol)*3+2+2] = dataBuf_[(row*dim+col)*3+2+2];
      }
   TRACE_DELETE_TAB( dataBuf_ );
   dataBuf_ = mirrTex;
}
#endif //TRANSFORM_TEXTURES

void WebNode::
uncompressDEM()
{
   ACE_TRACE( "(WebNode::uncompressTex)" );
   int dimy = (int)dataBuf_[0] * 256 + dataBuf_[1];
   int dimx = (int)dataBuf_[2] * 256 + dataBuf_[3];
   int elMin = (int)dataBuf_[4] * 256 + dataBuf_[5];
   int elMax = (int)dataBuf_[6] * 256 + dataBuf_[7];
   int elLevels = elMax - elMin + 1;
   int usedLength = dataSize_ - 8;
   
   float* demData;
   TRACE_NEW_2( demData, float[dimy*dimx] );
   unsigned char* compressedData = (unsigned char*)(dataBuf_ + 8);

   ewv_t* compressor = new_ewv(dimy, dimx, elMin, elMax, elLevels);

   int returnedQuality = dec_ewv(compressor,
				 compressedData,
				 &usedLength,
				 demData,
				 defaultDEMQuality,
				 /*scale*/1); //stink

   if(returnedQuality<0)
   {
      ACE_ERROR(( LM_ERROR, "[!](WebNode::uncompressTex) %s\n",
		  WN_ERROR[WN_EINVURL] ));
   }
   
   delete_ewv(&compressor);

   TRACE_DELETE_TAB( dataBuf_ );
   TRACE_NEW_2( dataBuf_, char[dimx*dimy*2+8] );
   dataSize_ = dimx*dimy*2+8;
   for(int i=0; i<dimy; i++)
      for(int j=0; j<dimx; j++)
      {
	 int demSampleValue = demData[i*dimx+j];
	 dataBuf_[8+(i*dimx+j)*2] = demSampleValue/256;
	 dataBuf_[8+(i*dimx+j)*2+1] = demSampleValue%256;
      }
   dataBuf_[0] = 0;
   dataBuf_[1] = 0;
   dataBuf_[2] = dimx/256;
   dataBuf_[3] = dimx%256;

   TRACE_DELETE_TAB( demData );
}

void WebNode::
processPPM()
{
   char* saveData = dataBuf_;

   //read "P6\n"
   dataBuf_ = dataBuf_ + 3;
   
   unsigned char c;
   unsigned char junk;
   int foundComment;
   int dimx = 0;
   int dimy = 0;
   
   do
   {
      c = dataBuf_[0];
      foundComment = c=='#';
      if(foundComment)
      {
	 do
	 {
	    dataBuf_ = dataBuf_ + 1;
	    junk = dataBuf_[0];
	 }
	 while(junk!='\n');
	 dataBuf_++;
      }
   }
   while(foundComment);
   if(dataBuf_[0] == '\n') dataBuf_++;
  
   while(dataBuf_[0]!=' ') {
      dimx = 10*dimx + dataBuf_[0] - '0'; 
      dataBuf_++;
   }
   dataBuf_++;
  
   while(dataBuf_[0]!='\n') {
      dimy = 10*dimy + dataBuf_[0] - '0'; 
      dataBuf_++;
   }
   dataBuf_++;

   while(dataBuf_[0]!='\n') dataBuf_++; //255\n
   dataBuf_++;

   char* parsedData;
   TRACE_NEW_2(parsedData, char[dimx*dimy*3+2]);
   parsedData[0] = dimy/256;
   parsedData[1] = dimy%256;
   memcpy(parsedData + 2, dataBuf_, dimx*dimy*3*sizeof(unsigned char));
   dataBuf_ = saveData;
   TRACE_DELETE_TAB(dataBuf_);
   dataBuf_ = parsedData;
   dataSize_ = 2+dimx*dimy*3*sizeof(unsigned char);
}
